@@ -18,6 +18,9 @@
 #include <osmocom/mgcp_client/mgcp_client.h>
 #include <osmocom/mgcp_client/mgcp_client_pool.h>
 
+#define STORE_UPTIME_INTERVAL	10 /* seconds */
+#define HNB_STORE_RAB_DURATIONS_INTERVAL 1 /* seconds */
+
 enum {
 	DMAIN,
 	DHNBAP,
@@ -70,15 +73,99 @@ enum hnb_ctrl_node {
 
 #define IUH_MSGB_SIZE	2048
 
+enum hnb_rate_ctr {
+	HNB_CTR_IUH_ESTABLISHED,
+	HNB_CTR_RANAP_PS_ERR_IND_UL,
+	HNB_CTR_RANAP_CS_ERR_IND_UL,
+	HNB_CTR_RANAP_PS_RESET_REQ_UL,
+	HNB_CTR_RANAP_CS_RESET_REQ_UL,
+
+	HNB_CTR_RANAP_PS_RAB_ACT_REQ,
+	HNB_CTR_RANAP_CS_RAB_ACT_REQ,
+
+	HNB_CTR_RANAP_PS_RAB_ACT_CNF,
+	HNB_CTR_RANAP_CS_RAB_ACT_CNF,
+
+	HNB_CTR_RANAP_PS_RAB_ACT_FAIL,
+	HNB_CTR_RANAP_CS_RAB_ACT_FAIL,
+
+	HNB_CTR_RANAP_PS_RAB_MOD_REQ,
+	HNB_CTR_RANAP_CS_RAB_MOD_REQ,
+
+	HNB_CTR_RANAP_PS_RAB_MOD_CNF,
+	HNB_CTR_RANAP_CS_RAB_MOD_CNF,
+
+	HNB_CTR_RANAP_PS_RAB_MOD_FAIL,
+	HNB_CTR_RANAP_CS_RAB_MOD_FAIL,
+
+	HNB_CTR_RANAP_PS_RAB_REL_REQ,
+	HNB_CTR_RANAP_CS_RAB_REL_REQ,
+
+	HNB_CTR_RANAP_PS_RAB_REL_CNF,
+	HNB_CTR_RANAP_CS_RAB_REL_CNF,
+
+	HNB_CTR_RANAP_PS_RAB_REL_FAIL,
+	HNB_CTR_RANAP_CS_RAB_REL_FAIL,
+
+	HNB_CTR_RANAP_PS_RAB_REL_IMPLICIT,
+	HNB_CTR_RANAP_CS_RAB_REL_IMPLICIT,
+
+	HNB_CTR_RUA_ERR_IND,
+
+	HNB_CTR_RUA_PS_CONNECT_UL,
+	HNB_CTR_RUA_CS_CONNECT_UL,
+
+	HNB_CTR_RUA_PS_DISCONNECT_UL,
+	HNB_CTR_RUA_CS_DISCONNECT_UL,
+	HNB_CTR_RUA_PS_DISCONNECT_DL,
+	HNB_CTR_RUA_CS_DISCONNECT_DL,
+
+	HNB_CTR_RUA_PS_DT_UL,
+	HNB_CTR_RUA_CS_DT_UL,
+	HNB_CTR_RUA_PS_DT_DL,
+	HNB_CTR_RUA_CS_DT_DL,
+
+	HNB_CTR_RUA_UDT_UL,
+	HNB_CTR_RUA_UDT_DL,
+
+	HNB_CTR_PS_PAGING_ATTEMPTED,
+	HNB_CTR_CS_PAGING_ATTEMPTED,
+
+	HNB_CTR_RAB_ACTIVE_MILLISECONDS_TOTAL,
+};
+
+enum hnb_stat {
+	HNB_STAT_UPTIME_SECONDS,
+};
+
 struct umts_cell_id {
-	uint16_t mcc;	/*!< Mobile Country Code */
-	uint16_t mnc;	/*!< Mobile Network Code */
-	uint16_t lac;	/*!< Locaton Area Code */
-	uint16_t rac;	/*!< Routing Area Code */
+	uint16_t mcc;	/*!< Mobile Country Code (0-999) */
+	uint16_t mnc;	/*!< Mobile Network Code (0-999) */
+	uint16_t lac;	/*!< Locaton Area Code (1-65534) */
+	uint16_t rac;	/*!< Routing Area Code (0-255) */
 	uint16_t sac;	/*!< Service Area Code */
 	uint32_t cid;	/*!< Cell ID */
 };
 const char *umts_cell_id_name(const struct umts_cell_id *ucid);
+int umts_cell_id_from_str(struct umts_cell_id *ucid, const char *instr);
+
+/*! are both given umts_cell_id euqal? */
+static inline bool umts_cell_id_equal(const struct umts_cell_id *a, const struct umts_cell_id *b)
+{
+	if (a->mcc != b->mcc)
+		return false;
+	if (a->mnc != b->mnc)
+		return false;
+	if (a->lac != b->lac)
+		return false;
+	if (a->rac != b->rac)
+		return false;
+	if (a->sac != b->sac)
+		return false;
+	if (a->cid != b->cid)
+		return false;
+	return true;
+}
 
 struct hnbgw_context_map;
 
@@ -152,6 +239,8 @@ struct hnbgw_cnpool {
 	struct rate_ctr_group *ctrs;
 };
 
+#define CNPOOL_CTR_INC(cnpool, x) rate_ctr_inc2((cnpool)->ctrs, x)
+
 /* A CN peer, like 'msc 0' or 'sgsn 23' */
 struct hnbgw_cnlink {
 	struct llist_head entry;
@@ -189,6 +278,8 @@ struct hnbgw_cnlink {
 #define LOG_CNLINK(CNLINK, SUBSYS, LEVEL, FMT, ARGS...) \
 	LOGP(SUBSYS, LEVEL, "(%s) " FMT, (CNLINK) ? (CNLINK)->name : "null", ##ARGS)
 
+#define CNLINK_CTR_INC(cnlink, x) rate_ctr_inc2((cnlink)->ctrs, x)
+
 struct hnbgw_cnlink *cnlink_get_nr(struct hnbgw_cnpool *cnpool, int nr, bool create_if_missing);
 
 static inline bool cnlink_is_cs(const struct hnbgw_cnlink *cnlink)
@@ -233,6 +324,35 @@ struct hnb_context {
 
 	/* linked list of hnbgw_context_map */
 	struct llist_head map_list;
+
+	/*! pointer to the associated hnb persistent state. Always present after HNB-Register */
+	struct hnb_persistent *persistent;
+};
+
+#define HNBP_CTR(hnbp, x) rate_ctr_group_get_ctr((hnbp)->ctrs, x)
+#define HNBP_CTR_INC(hnbp, x) rate_ctr_inc(HNBP_CTR(hnbp, x))
+#define HNBP_CTR_ADD(hnbp, x, y) rate_ctr_add2((hnbp)->ctrs, x, y)
+
+#define HNBP_STAT(hbp, x) osmo_stat_item_group_get_item((hnbp)->statg, x)
+#define HNBP_STAT_SET(hnbp, x, val) osmo_stat_item_set(HNBP_STAT(hnbp, x), val)
+
+/* persistent data for one HNB.  This continues to exist even as conn / hnb_context is deleted on disconnect */
+struct hnb_persistent {
+	/*! Entry in HNBGW-global list of hnb_persistent */
+	struct llist_head list;
+	/*! back-pointer to hnb_context.  Can be NULL if no context at this point */
+	struct hnb_context *ctx;
+
+	/*! unique cell identity; copied from HNB REGISTER REQ */
+	struct umts_cell_id id;
+	/*! stringified version of the cell identiy above (for printing/naming) */
+	const char *id_str;
+
+	/*! copied from HNB-Identity-Info IE */
+	time_t updowntime;
+
+	struct rate_ctr_group *ctrs;
+	struct osmo_stat_item_group *statg;
 };
 
 struct ue_context {
@@ -259,18 +379,26 @@ struct hnbgw {
 		bool hnbap_allow_tmsi;
 		/*! print hnb-id (true) or MCC-MNC-LAC-RAC-SAC (false) in logs */
 		bool log_prefix_hnb_id;
+		bool accept_all_hnb;
 		struct mgcp_client_conf *mgcp_client;
 		struct {
 			char *local_addr;
 			uint16_t local_port;
 			char *remote_addr;
 			uint16_t remote_port;
+			struct {
+				char *access;
+				char *core;
+			} netinst;
 		} pfcp;
 	} config;
 	/*! SCTP listen socket for incoming connections */
 	struct osmo_stream_srv_link *iuh;
 	/* list of struct hnb_context */
 	struct llist_head hnb_list;
+	/* list of struct hnb_persistent */
+	struct llist_head hnb_persistent_list;
+	struct osmo_timer_list store_uptime_timer;
 	/* list of struct ue_context */
 	struct llist_head ue_list;
 	/* next availble UE Context ID */
@@ -294,6 +422,8 @@ struct hnbgw {
 		struct osmo_pfcp_endpoint *ep;
 		struct osmo_pfcp_cp_peer *cp_peer;
 	} pfcp;
+
+	struct osmo_timer_list hnb_store_rab_durations_timer;
 };
 
 extern struct hnbgw *g_hnbgw;
@@ -319,6 +449,10 @@ void ue_context_free(struct ue_context *ue);
 void hnb_context_release(struct hnb_context *ctx);
 void hnb_context_release_ue_state(struct hnb_context *ctx);
 
+struct hnb_persistent *hnb_persistent_alloc(const struct umts_cell_id *id);
+struct hnb_persistent *hnb_persistent_find_by_id(const struct umts_cell_id *id);
+void hnb_persistent_free(struct hnb_persistent *hnbp);
+
 void hnbgw_vty_init(void);
 int hnbgw_vty_go_parent(struct vty *vty);
 
@@ -334,4 +468,6 @@ static inline bool hnb_gw_is_gtp_mapping_enabled(void)
 
 struct msgb *hnbgw_ranap_msg_alloc(const char *name);
 
-int hnbgw_peek_l3(struct hnbgw_context_map *map, struct msgb *ranap_msg);
+int hnbgw_peek_l3_ul(struct hnbgw_context_map *map, struct msgb *ranap_msg);
+
+unsigned long long hnb_get_updowntime(const struct hnb_context *ctx);
